from unittest.mock import patch

import pytest
from django.contrib.admin import AdminSite
from django.contrib.auth import get_user_model
from django.contrib.messages import get_messages
from dynamic_preferences.registries import global_preferences_registry
from rest_framework import status
from wbcore.contrib.authentication.factories import UserFactory

from wbcrm.synchronization.activity.admin import UserSyncAdmin

from .fixtures import TestOutlookSyncFixture, msgraph_fixture


@pytest.mark.parametrize(
    "backend, credentials, subscription, tenant_id",
    [
        (False, True, {"id": "fake_id"}, "fake_tenant_id"),
        (True, False, None, None),
        (True, True, None, None),
        (True, True, {"id": "fake_id"}, None),
        (True, True, {"id": "fake_id"}, "fake_tenant_id"),
    ],
)
@patch("wbcrm.synchronization.activity.backends.outlook.backend.MicrosoftGraphAPI")
@pytest.mark.django_db
class TestAdminUserWebhook(TestOutlookSyncFixture):
    def _initialiation(self, mock_msgraph, backend, credentials, subscription, tenant_id, credentials_fixture):
        preferences = global_preferences_registry.manager()
        preferences["wbactivity_sync__sync_backend_calendar"] = preferences[
            "wbactivity_sync__outlook_sync_credentials"
        ] = ""
        if backend:
            preferences["wbactivity_sync__sync_backend_calendar"] = (
                "wbcrm.synchronization.activity.backends.outlook.backend.OutlookSyncBackend"
            )
        mock_msgraph.return_value.status_code = status.HTTP_200_OK
        mock_msgraph.return_value = msgraph_fixture()
        if credentials:
            preferences["wbactivity_sync__outlook_sync_credentials"] = credentials_fixture
            msgraph_fixture._subscription = subscription
            msgraph_fixture.tenant_id = tenant_id
        else:
            mock_msgraph.side_effect = AssertionError("Invalid URL")
        user = UserFactory(is_superuser=True)
        if subscription:
            user.metadata = {"outlook": {"subscription": subscription}}
            user.save()
        return user

    def test_set_web_hook(
        self, mock_msgraph, backend, credentials, subscription, tenant_id, fixture_request, credentials_fixture
    ):
        user = self._initialiation(mock_msgraph, backend, credentials, subscription, tenant_id, credentials_fixture)
        fixture_request.user = user
        assert len([m.message for m in get_messages(fixture_request)]) == 0
        user_admin = UserSyncAdmin(UserFactory, AdminSite())
        user_admin.set_web_hook(fixture_request, get_user_model().objects.filter(id=user.id))
        messages = [m.message for m in get_messages(fixture_request)]
        assert len(messages) == 1
        if backend:
            if credentials:
                if subscription or tenant_id:
                    assert "Operation completed" in messages[0]
                else:
                    assert messages[0] == f"Operation Failed, Outlook TenantId not found for: {user}"
            else:
                assert "Operation Failed, Invalid URL" in messages[0]
        else:
            assert messages[0] == "Operation Failed, No backend set in preferences"

    def test_stop_web_hook(
        self, mock_msgraph, backend, credentials, subscription, tenant_id, fixture_request, credentials_fixture
    ):
        user = self._initialiation(mock_msgraph, backend, credentials, subscription, tenant_id, credentials_fixture)
        fixture_request.user = user
        assert len([m.message for m in get_messages(fixture_request)]) == 0
        user_admin = UserSyncAdmin(UserFactory, AdminSite())
        user_admin.stop_web_hook(fixture_request, get_user_model().objects.filter(id=user.id))
        messages = [m.message for m in get_messages(fixture_request)]
        assert len(messages) == 1
        if backend:
            if subscription:
                if credentials:
                    assert "Operation completed" in messages[0]
                else:
                    assert "Operation Failed, Invalid URL" in messages[0]
            else:
                assert f"Operation Failed, {user} has no active webhook"
        else:
            assert messages[0] == "Operation Failed, No backend set in preferences"

    def test_check_web_hook(
        self, mock_msgraph, backend, credentials, subscription, tenant_id, fixture_request, credentials_fixture
    ):
        user = self._initialiation(mock_msgraph, backend, credentials, subscription, tenant_id, credentials_fixture)
        fixture_request.user = user
        assert len([m.message for m in get_messages(fixture_request)]) == 0
        user_admin = UserSyncAdmin(UserFactory, AdminSite())
        user_admin.check_web_hook(fixture_request, get_user_model().objects.filter(id=user.id))
        messages = [m.message for m in get_messages(fixture_request)]
        assert len(messages) == 1
        if backend:
            if credentials:
                if subscription:
                    assert "Operation completed" in messages[0]
                elif tenant_id:
                    assert (
                        "Operation Failed, Webhook not found. Number of subscriptions found in outlook for"
                        in messages[0]
                    )
                else:
                    assert messages[0] == f"Operation Failed, Webhook not found. TenantId not found for {user}"
            else:
                assert "Operation Failed, Invalid URL" in messages[0]
        else:
            assert messages[0] == "Operation Failed, No backend set in preferences"



from collections import OrderedDict
from unittest import TestCase

import pytest
from django.core.exceptions import ValidationError as DjangoValidationError
from rest_framework.exceptions import ValidationError
from rest_framework.test import APIRequestFactory
from wbcore.contrib.authentication.factories import UserFactory
from wbcore.contrib.directory.factories import PersonFactory
from wbcore.contrib.directory.models import Entry

from wbcrm.factories import ActivityFactory, GroupFactory, ProductFactory
from wbcrm.serializers import ActivityModelSerializer, ProductModelSerializer
from wbcrm.serializers.activities import handle_representation


@pytest.mark.django_db
class TestActivitySerializers(TestCase):
    @pytest.mark.unit
    def test_remove_group_member_from_instance(self):
        group_member_a = PersonFactory()
        group_member_a_entry = Entry.objects.get(id=group_member_a.id)
        group_member_b = PersonFactory()
        group_member_b_entry = Entry.objects.get(id=group_member_b.id)
        group = GroupFactory(members=(group_member_a_entry, group_member_b_entry))
        activity = ActivityFactory(groups=(group,), participants=(group_member_a, group_member_b))

        request = APIRequestFactory().get("")
        request.user = UserFactory()
        request.parser_context = {}
        serializer = ActivityModelSerializer(activity, context={"request": request})
        new_data = {
            "participants": [group_member_b],
        }
        assert serializer.data
        assert group_member_a.id in serializer.data["participants"]
        assert group_member_b.id in serializer.data["participants"]
        with self.assertRaises(ValidationError):
            serializer.validate(new_data)

    @pytest.mark.unit
    def test_remove_all_group_members_from_instance(self):
        group_member_a = PersonFactory()
        group_member_a_entry = Entry.objects.get(id=group_member_a.id)
        group_member_b = PersonFactory()
        group_member_b_entry = Entry.objects.get(id=group_member_b.id)
        group = GroupFactory(members=(group_member_a_entry, group_member_b_entry))
        activity = ActivityFactory(groups=(group,), participants=(group_member_a, group_member_b))

        request = APIRequestFactory().get("")
        request.user = UserFactory()
        request.parser_context = {}
        serializer = ActivityModelSerializer(activity, context={"request": request})
        new_data = {
            "participants": [],
        }
        assert serializer.data
        assert set(serializer.data["participants"]) == {group_member_a.id, group_member_b.id}
        with self.assertRaises(ValidationError):
            serializer.validate(new_data)

    @pytest.mark.unit
    def test_remove_all_group_members_and_groups_from_instance(self):
        group_member_a = PersonFactory()
        group_member_a_entry = Entry.objects.get(id=group_member_a.id)
        group_member_b = PersonFactory()
        group_member_b_entry = Entry.objects.get(id=group_member_b.id)
        group = GroupFactory(members=(group_member_a_entry, group_member_b_entry))
        activity = ActivityFactory(groups=(group,), participants=(group_member_a, group_member_b))

        request = APIRequestFactory().get("")
        request.user = UserFactory()
        request.parser_context = {}
        serializer = ActivityModelSerializer(activity, context={"request": request})
        new_data = {
            "participants": [],
            "groups": [],
        }
        assert serializer.data
        try:
            validated_data = serializer.validate(new_data)
        except DjangoValidationError:
            self.fail("Activity threw error in validation method!")
        else:
            self.assertEqual(validated_data, new_data)

    @pytest.mark.unit
    def test_remove_not_group_membersfrom_instance(self):
        group_member_a = PersonFactory()
        group_member_a_entry = Entry.objects.get(id=group_member_a.id)
        not_group_member = PersonFactory()
        not_group_member_b_entry = Entry.objects.get(id=not_group_member.id)
        group = GroupFactory(members=(group_member_a_entry, not_group_member_b_entry))
        activity = ActivityFactory(groups=(group,), participants=(group_member_a, not_group_member))

        request = APIRequestFactory().get("")
        request.user = UserFactory()
        request.parser_context = {}
        serializer = ActivityModelSerializer(activity, context={"request": request})
        new_data = {
            "participants": [
                group_member_a.id,
            ],
            "groups": [],
        }
        assert serializer.data
        try:
            validated_data = serializer.validate(new_data)
        except DjangoValidationError:
            self.fail("Activity threw error in validation method!")
        else:
            self.assertEqual(validated_data, new_data)


@pytest.mark.django_db
class TestActivitySerializersHelperFunction:
    @pytest.mark.parametrize("is_private, is_confidential", [(True, False), (True, False)])
    def test_handle_representaion(self, is_private, is_confidential):
        test_representation = OrderedDict(
            [
                ("id", "Test-ID"),
                ("title", "Test-Title"),
                ("is_private", is_private),
                ("is_confidential", is_confidential),
                ("foo", "Foo"),
                ("bar", "Bar"),
            ]
        )
        test_representation = handle_representation(test_representation)
        assert test_representation["id"] == "Test-ID"
        if is_private:
            assert test_representation["title"] == "Private Activity"
            assert test_representation["foo"] is None
            assert test_representation["bar"] is None
        elif not is_private and is_confidential:
            assert test_representation["title"] == "Confidential Activity"
            assert test_representation["foo"] is None
            assert test_representation["bar"] is None
        elif not is_private and not is_confidential:
            assert test_representation["title"] == "Test-Title"
            assert test_representation["foo"] == "Foo"
            assert test_representation["bar"] == "Bar"


@pytest.mark.django_db
class TestUtilsSerializers:
    def test_duplicate_company_product(self):
        product = ProductFactory(is_competitor=False)
        serializer = ProductModelSerializer()
        data = {"title": product.title, "is_competitor": product.is_competitor}
        with pytest.raises(DjangoValidationError):
            serializer.validate(data)

    def test_non_duplicate_competitors_product(self):
        product = ProductFactory(is_competitor=False)
        serializer = ProductModelSerializer()
        data = {"title": product.title, "is_competitor": not product.is_competitor}
        assert serializer.validate(data)

    def test_non_duplicate_company_product(self):
        product = ProductFactory(is_competitor=True)
        serializer = ProductModelSerializer()
        data = {"title": product.title, "is_competitor": not product.is_competitor}
        assert serializer.validate(data)

    def test_duplicate_competitors_product(self):
        product = ProductFactory(is_competitor=True)
        serializer = ProductModelSerializer()
        data = {"title": product.title, "is_competitor": product.is_competitor}
        with pytest.raises(DjangoValidationError):
            serializer.validate(data)



from django.db.models import Model
from django.forms import ValidationError
from django.utils.translation import gettext
from rest_framework import serializers
from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers

from wbcrm.models.recurrence import Recurrence


class RecurrenceModelSerializerMixin:
    @wb_serializers.register_only_instance_resource()
    def next_occurrence(self, instance: Model, request, user, **kwargs) -> dict:
        resources = {}
        if next_occurrence := instance.next_occurrence:
            resources["next_occurrence"] = reverse(
                f"{instance.get_endpoint_basename()}-detail", args=[next_occurrence.id], request=request
            )
        return resources

    @wb_serializers.register_only_instance_resource()
    def previous_occurrence(self, instance: Model, request, user, **kwargs) -> dict:
        resources = {}
        if previous_occurrence := instance.previous_occurrence:
            resources["previous_occurrence"] = reverse(
                f"{instance.get_endpoint_basename()}-detail", args=[previous_occurrence.id], request=request
            )
        return resources

    @wb_serializers.register_only_instance_resource()
    def get_parent_occurrence(self, instance: Model, request, user, **kwargs) -> dict:
        resources = {}
        if instance.parent_occurrence:
            resources["get_parent_occurrence"] = reverse(
                f"{instance.get_endpoint_basename()}-detail", args=[instance.parent_occurrence.id], request=request
            )
        return resources

    @wb_serializers.register_only_instance_resource()
    def delete_occurrences(self, instance, request, user, **kwargs):
        resources = dict()
        parent = instance.parent_occurrence if instance.parent_occurrence else instance
        child_occurrences = parent.get_recurrent_valid_children()
        if instance.period and child_occurrences.exists():
            child_occurrences = child_occurrences.filter(period__startswith__gt=instance.period.lower)
        if child_occurrences.exists():
            resources["delete_next_occurrences"] = reverse(
                f"{instance.get_endpoint_basename()}-delete-next-occurrences", args=[instance.id], request=request
            )
        return resources

    def validate(self, data):
        period = data.get("period", self.instance.period if self.instance else None)
        recurrence_count = data.get("recurrence_count", self.instance.recurrence_count if self.instance else None)
        recurrence_end = data.get("recurrence_end", self.instance.recurrence_end if self.instance else None)
        repeat_choice = data.get("repeat_choice", self.instance.repeat_choice if self.instance else None)
        if not period:
            raise serializers.ValidationError({"period": gettext("Please provide a valid timeframe.")})

        if recurrence_end and recurrence_count:
            error = gettext("You can only pick either a recurrence count or an end date but not both.")
            raise ValidationError({"recurrence_end": error, "recurrence_count": error})

        if (
            not self.instance
            and (recurrence_end or recurrence_count)
            and repeat_choice == Recurrence.ReoccuranceChoice.NEVER
        ):
            data["recurrence_end"] = None
            data["recurrence_count"] = None
        if data.get("repeat_choice") and data.get("repeat_choice") != Recurrence.ReoccuranceChoice.NEVER:
            if data.get("recurrence_end") and period.lower.date() >= data.get("recurrence_end"):
                raise ValidationError(
                    {
                        "recurrence_end": gettext(
                            'The "Repeat Until" date needs to be after the "Recurrence Start" date.'
                        )
                    }
                )
            if data.get("repeat_choice") == Recurrence.ReoccuranceChoice.BUSINESS_DAILY and period.lower.weekday() > 4:
                raise ValidationError({"period": gettext("Period must correspond to the recurrence 'Business Daily'")})

        if self.instance and self.instance.period and self.instance.is_recurrent:
            if (
                self.instance.period.lower.date() != period.lower.date()
                or self.instance.period.upper.date() != period.upper.date()
            ):
                raise ValidationError(
                    {"period": gettext("It is only possible to change the time of the period of an occurrence.")}
                )
        return super().validate(data)



import calendar as calendar_reference
from datetime import datetime
from typing import Any

import pandas as pd
import pytz
from dateutil import parser, rrule
from django.utils.timezone import get_default_timezone, get_default_timezone_name
from psycopg.types.range import TimestamptzRange
from wbcore.utils.rrules import (
    convert_rrulestr_to_dict,
    convert_weekday_rrule_to_day_name,
)

from wbcrm.typings import Activity as ActivityDTO
from wbcrm.typings import ConferenceRoom as ConferenceRoomDTO
from wbcrm.typings import ParticipantStatus as ParticipantStatusDTO
from wbcrm.typings import Person as PersonDTO


class OutlookParser:
    default_visibility = "normal"
    visibility_map = {
        ActivityDTO.Visibility.PUBLIC.name: default_visibility,
        ActivityDTO.Visibility.PRIVATE.name: "private",
        ActivityDTO.Visibility.CONFIDENTIAL.name: "confidential",
    }
    default_response_status = "notResponded"
    response_status_map = {
        ParticipantStatusDTO.ParticipationStatus.NOTRESPONDED.name: default_response_status,
        ParticipantStatusDTO.ParticipationStatus.ATTENDS.name: "accepted",
        ParticipantStatusDTO.ParticipationStatus.CANCELLED.name: "declined",
        ParticipantStatusDTO.ParticipationStatus.MAYBE.name: "tentativelyAccepted",
    }
    default_reminder_minutes = 15
    reminder_map = {
        ActivityDTO.ReminderChoice.NEVER.name: -1,
        ActivityDTO.ReminderChoice.EVENT_TIME.name: 0,
        ActivityDTO.ReminderChoice.MINUTES_5.name: 5,
        ActivityDTO.ReminderChoice.MINUTES_15.name: default_reminder_minutes,
        ActivityDTO.ReminderChoice.MINUTES_30.name: 30,
        ActivityDTO.ReminderChoice.HOURS_1.name: 60,
        ActivityDTO.ReminderChoice.HOURS_2.name: 120,
        ActivityDTO.ReminderChoice.HOURS_12.name: 720,
        ActivityDTO.ReminderChoice.WEEKS_1.name: 10080,
    }

    @classmethod
    def convert_visibility_to_sensitivity(cls, visibility: str) -> str:
        return cls.visibility_map.get(visibility, cls.default_visibility)

    @classmethod
    def convert_sensitivity_to_visibility(cls, sensitivity: str) -> str:
        visibility_map_inv = {v: k for k, v in cls.visibility_map.items()}
        return visibility_map_inv.get(sensitivity, visibility_map_inv[cls.default_visibility])

    @classmethod
    def convert_participant_status_to_attendee_status(cls, participant_status: str) -> str:
        if participant_status == ParticipantStatusDTO.ParticipationStatus.ATTENDS_DIGITALLY.name:
            return "accepted"
        if participant_status == ParticipantStatusDTO.ParticipationStatus.PENDING_INVITATION.name:
            return cls.default_response_status
        return cls.response_status_map.get(participant_status, cls.default_response_status)

    @classmethod
    def convert_attendee_status_to_participant_status(cls, attendee_status: str) -> str:
        if attendee_status == "organizer":
            return ParticipantStatusDTO.ParticipationStatus.ATTENDS.name
        response_status_map_inv = {v: k for k, v in cls.response_status_map.items()}
        return response_status_map_inv.get(attendee_status, response_status_map_inv[cls.default_response_status])

    @classmethod
    def convert_reminder_choice_to_minutes(cls, choice: str) -> int:
        return cls.reminder_map[choice]

    @classmethod
    def convert_reminder_minutes_to_choice(cls, minutes: int) -> str:
        reminder_map_inv = {v: k for k, v in cls.reminder_map.items()}
        return reminder_map_inv.get(minutes, reminder_map_inv[cls.default_reminder_minutes])

    @classmethod
    def convert_string_to_datetime(cls, date_time_str: str, time_zone: str) -> datetime:
        date_time = parser.parse(date_time_str)
        try:
            return pytz.timezone(time_zone).localize(date_time)
        except pytz.exceptions.UnknownTimeZoneError:
            return pytz.timezone(get_default_timezone_name()).localize(date_time)

    @classmethod
    def convert_to_all_day_period(cls, period: TimestamptzRange) -> TimestamptzRange:
        return TimestamptzRange(
            lower=datetime(
                year=period.lower.year, month=period.lower.month, day=period.lower.day, tzinfo=get_default_timezone()
            ),
            upper=datetime(
                year=period.upper.year, month=period.upper.month, day=period.upper.day, tzinfo=get_default_timezone()
            ),
        )

    @classmethod
    def deserialize_person(cls, email: str, mame: str) -> PersonDTO:
        if (items := mame.split(" ")) and len(items) > 1:
            first_name = items[0]
            last_name = " ".join(items[1:])
        else:
            last_name = first_name = items[0]
        return PersonDTO(first_name=first_name, last_name=last_name, email=email)

    @classmethod
    def deserialize_conference_room(cls, email: str, name: str) -> ConferenceRoomDTO:
        return ConferenceRoomDTO(name=name, email=email)

    @classmethod
    def deserialize_participants(cls, event: dict) -> tuple[list[ParticipantStatusDTO], list[ConferenceRoomDTO]]:
        participants = {
            event["organizer.email_address.address"]: ParticipantStatusDTO(
                person=OutlookParser.deserialize_person(
                    event["organizer.email_address.address"], event["organizer.email_address.name"]
                ),
                status=ParticipantStatusDTO.ParticipationStatus.ATTENDS.name,
            )
        }
        conference_rooms = {}
        for attendee in event.get("attendees", []):
            if mail_info := attendee.get("emailAddress"):
                email = mail_info["address"]
                name = mail_info.get("name", email)
                if attendee.get("type") == "resource":
                    if not conference_rooms.get(email):
                        conference_rooms[email] = cls.deserialize_conference_room(email, name)
                else:
                    if (status_participant := attendee["status"].get("response")) and (
                        event["is_organizer"] or (not event["is_organizer"] and status_participant != "none")
                    ):
                        status = cls.convert_attendee_status_to_participant_status(status_participant)
                    else:
                        status = None
                    defaults = {"status": status} if status else {}
                    if not participants.get(email):
                        participants[email] = ParticipantStatusDTO(
                            person=cls.deserialize_person(email, name), **defaults
                        )
        # handle conference room when user adds conference room from participant list
        for location in event.get("locations", []):
            if location.get("locationType") == "conferenceRoom":
                email = location["locationUri"]
                name = location.get("displayName", email)
                conference_rooms[email] = cls.deserialize_conference_room(email, name)
        return list(participants.values()), list(conference_rooms.values())

    @classmethod
    def serialize_person(cls, person: PersonDTO) -> dict:
        return {
            "emailAddress": {
                "address": person.email,
                "name": f"{person.first_name} {person.last_name}",
            }
        }

    @classmethod
    def serialize_conference_room(cls, conference_room_dto=ConferenceRoomDTO) -> dict:
        return {
            "emailAddress": {"address": conference_room_dto.email, "name": conference_room_dto.name},
            "type": "resource",
        }

    @classmethod
    def serialize_participants(cls, participants_dto: list[ParticipantStatusDTO]) -> list[dict[str, Any]]:
        attendees = []
        for participant_dto in participants_dto:
            status = cls.convert_participant_status_to_attendee_status(participant_dto.status)
            _data = {
                **cls.serialize_person(participant_dto.person),
                "type": "optional",
            }
            if status != "notResponded":
                _data["status"] = {"response": status}
            attendees += [_data]
        return attendees

    @classmethod
    def _convert_weekday_to_day_name(cls, weekday: rrule.weekday) -> str:
        default_day_name = calendar_reference.day_name[0]  # Monday
        if weekday and (_day_name := convert_weekday_rrule_to_day_name(weekday)):
            return _day_name
        return default_day_name

    @classmethod
    def deserialize_recurring_activities(cls, event: dict) -> dict:
        recurrence = {
            "recurrence_end": None,
            "recurrence_count": 0,
            "repeat_choice": ActivityDTO.ReoccuranceChoice.NEVER.value,
        }
        if event.get("type") == "seriesMaster":
            freq = event["recurrence.pattern.type"]
            interval = event["recurrence.pattern.interval"]
            if freq == "yearly":
                recurrence["repeat_choice"] = ActivityDTO.ReoccuranceChoice.YEARLY.value
            elif freq == "monthly":
                if interval == 3:
                    recurrence["repeat_choice"] = ActivityDTO.ReoccuranceChoice.QUARTERLY.value
                else:
                    recurrence["repeat_choice"] = ActivityDTO.ReoccuranceChoice.MONTHLY.value
            elif freq == "weekly":
                days_of_week = {_day.lower() for _day in event.get("recurrence.pattern.days_of_week", [])}
                if interval == 2:
                    recurrence["repeat_choice"] = ActivityDTO.ReoccuranceChoice.BIWEEKLY.value
                elif interval == 1 and days_of_week == {"monday", "tuesday", "wednesday", "thursday", "friday"}:
                    recurrence["repeat_choice"] = ActivityDTO.ReoccuranceChoice.BUSINESS_DAILY.value
                else:
                    recurrence["repeat_choice"] = ActivityDTO.ReoccuranceChoice.WEEKLY.value
            else:
                recurrence["repeat_choice"] = ActivityDTO.ReoccuranceChoice.DAILY.value
            if event["recurrence.range.type"] == "endDate":
                recurrence["recurrence_end"] = datetime.strptime(event["recurrence.range.end_date"], "%Y-%m-%d")
            else:
                recurrence["recurrence_count"] = event["recurrence.range.number_of_occurrences"] - 1
        return recurrence

    @classmethod
    def serialize_recurring_activities(cls, activity_dto: ActivityDTO) -> dict:
        recurrence = {}
        if activity_dto.is_recurrent and (rule_dict := convert_rrulestr_to_dict(activity_dto.repeat_choice)):
            freq_names = {idx: freq for idx, freq in enumerate(rrule.FREQNAMES)}
            if freq_name := freq_names.get(rule_dict.get("freq")):
                pattern = {
                    "interval": rule_dict.get("interval", 1),
                    "firstDayOfWeek": cls._convert_weekday_to_day_name(rule_dict.get("wkst")),
                }
                if freq_name == "MONTHLY":
                    pattern.update({"type": "absoluteMonthly", "dayOfMonth": activity_dto.period.lower.day})
                elif freq_name == "YEARLY":
                    pattern.update(
                        {
                            "type": "absoluteYearly",
                            "month": activity_dto.period.lower.month,
                            "dayOfMonth": activity_dto.period.lower.day,
                        }
                    )
                else:  # daily or weekly
                    pattern["type"] = freq_name.lower()
                    if weekdays := rule_dict.get("byweekday", []):
                        pattern.update(
                            {
                                "type": "weekly",
                                "daysOfWeek": [cls._convert_weekday_to_day_name(weekday) for weekday in weekdays],
                            }
                        )
                    elif freq_name == "WEEKLY":
                        pattern["daysOfWeek"] = [calendar_reference.day_name[activity_dto.period.lower.weekday()]]

                range = {"startDate": activity_dto.period.lower.strftime("%Y-%m-%d")}
                if activity_dto.recurrence_count:
                    range.update(
                        {
                            "type": "numbered",
                            "numberOfOccurrences": activity_dto.recurrence_count + 1,
                            "recurrenceTimeZone": activity_dto.period.lower.tzname(),
                        }
                    )
                else:
                    range.update(
                        {
                            "type": "endDate",
                            "endDate": activity_dto.recurrence_end.strftime("%Y-%m-%d"),
                        }
                    )

                recurrence = {"recurrence": {"pattern": pattern, "range": range}}
        return recurrence

    @classmethod
    def serialize_recurrence_range_with_end_date(cls, recurrence: dict, end_date: datetime) -> dict:
        if (
            (recc := recurrence.get("recurrence"))
            and (_range := recc.get("range"))
            and (start_date := _range.get("startDate"))
        ):
            if start_date <= end_date.strftime("%Y-%m-%d"):
                recurrence["recurrence"]["range"] = {
                    "type": "endDate",
                    "startDate": start_date,
                    "endDate": end_date.strftime("%Y-%m-%d"),
                }
                if tz := _range.get("recurrenceTimeZone"):
                    recurrence["recurrence"]["range"]["recurrenceTimeZone"] = tz
                return recurrence

    @classmethod
    def serialize_event_keys(cls, event: dict) -> dict:
        inv_map = {value: key for key, value in _map.items()}
        return dict(map(lambda item: (inv_map.get(item[0], item[0]), item[1]), event.items()))


_map = {
    "lastModifiedDateTime": "last_modified_date_time",
    "startDateTime": "start_date_time",
    "endDateTime": "end_date_time",
    "organizer": "organizer",
    "displayName": "display_name",
    "businessPhones": "business_phones",
    "mobilePhone": "mobile_phone",
    "surname": "surname",
    "givenName": "given_name",
    "mailNickname": "mail_nickname",
    "userPrincipalName": "user_principal_name",
    "notificationUrl": "notification_url",
    "changeType": "change_type",
    "expirationDateTime": "expiration_date_time",
    "resourceData": "resource_data",
    "subscriptionExpirationDateTime": "subscription_expiration_date_time",
    "clientState": "client_state",
    "tenantId": "tenant_id",
    "subscriptionId": "subscription_id",
    "organizer.user": "organizer.user",
    "organizer.user.id": "organizer.user.id",
    "organizer.user.displayName": "organizer.user.display_name",
    "organizer.phone": "organizer.phone",
    "organizer.phone.id": "organizer.phone.id",
    "organizer.phone.displayName": "organizer.phone.display_name",
    "organizer.guest": "organizer.guest",
    "organizer.guest.id": "organizer.guest.id",
    "organizer.guest.displayName": "organizer.guest.display_name",
    "organizer.spoolUser": "organizer.splool_user",
    "organizer.acsUser": "organizer.acs_user",
    "organizer.encrypted": "organizer.encrypted",
    "organizer.onPremises": "organizer.on_premises",
    "organizer.acsApplicationInstance": "organizer.acs_application_instance",
    "organizer.spoolApplicationInstance": "organizer.spool_application_instance",
    "organizer.applicationInstance": "organizer.application_instance",
    "organizer.application": "organizer.application",
    "organizer.device": "organizer.device",
    "joinWebUrl": "join_web_url",
    "preferredLanguage": "preferred_language",
    "officeLocation": "office_location",
    "jobTitle": "job_title",
    "createdDateTime": "created_date_time",
    "isReminderOn": "is_reminder_on",
    "reminderMinutesBeforeStart": "reminder_minutes_before_start",
    "hasAttachments": "has_attachments",
    "webLink": "web_link",
    "start.dateTime": "start.date_time",
    "start.timeZone": "start.time_zone",
    "end.dateTime": "end.date_time",
    "end.timeZone": "end.time_zone",
    "location.displayName": "location.display_name",
    "location.locationUri": "location.location_uri",
    "location.locationType": "location.location_type",
    "location.uniqueId": "location.unique_id",
    "location.uniqueIdType": "location.unique_id_type",
    "location.address.type": "location.address.type",
    "location.address.street": "location.address.street",
    "location.address.city": "location.address.city",
    "location.address.postalCode": "location.address.postal_code",
    "location.address.countryOrRegion": "location.address.country_or_region",
    "recurrence.pattern.daysOfWeek": "recurrence.pattern.days_of_week",
    "recurrence.pattern.dayOfMonth": "recurrence.pattern.day_of_month",
    "recurrence.pattern.firstDayOfWeek": "recurrence.pattern.first_day_of_week",
    "recurrence.range.recurrenceTimeZone": "recurrence.range.recurrence_time_zone",
    "recurrence.range.numberOfOccurrences": "recurrence.range.number_of_occurrences",
    "recurrence.range.startDate": "recurrence.range.start_date",
    "recurrence.range.endDate": "recurrence.range.end_date",
    "organizer.emailAddress.name": "organizer.email_address.name",
    "organizer.emailAddress.address": "organizer.email_address.address",
    "responseStatus.response": "response_status.response",
    "responseStatus.time": "response_status.time",
    "attendees": "attendees",
    "@odata.context": "odata_context",
    "@odata.etag": "odata_etag",
    "applicationId": "application_id",
    "includeResourceData": "include_resource_data",
    "latestSupportedTlsVersion": "latest_supported_tls_version",
    "encryptionCertificate": "encryption_certificate",
    "encryptionCertificateId": "encryption_certificate_id",
    "notificationQueryOptions": "notification_query_options",
    "notificationContentType": "notification_content_type",
    "lifecycleNotificationUrl": "lifecycle_notification_url",
    "creatorId": "creator_id",
    "isDefaultCalendar": "is_default_calendar",
    "hexColor": "hex_color",
    "changeKey": "change_key",
    "canShare": "can_share",
    "canViewPrivateItems": "can_view_private_items",
    "isShared": "is_shared",
    "isSharedWithMe": "is_shared_with_me",
    "canEdit": "can_edit",
    "allowedOnlineMeetingProviders": "allowed_online_meeting_providers",
    "defaultOnlineMeetingProvider": "default_online_meeting_provider",
    "isTallyingResponses": "is_tallying_responses",
    "isRemovable": "is_removable",
    "teamsForBusiness": "teams_for_business",
    "transactionId": "transaction_id",
    "originalStartTimeZone": "original_start_time_zone",
    "originalEndTimeZone": "original_end_time_zone",
    "bodyPreview": "body_preview",
    "isAllDay": "is_all_day",
    "isCancelled": "is_cancelled",
    "isOrganizer": "is_organizer",
    "responseRequested": "response_requested",
    "seriesMasterId": "series_master_id",
    "showAs": "show_as",
    "onlineMeeting": "online_meeting",
    "onlineMeetingUrl": "online_meeting_url",
    "isOnlineMeeting": "is_online_meeting",
    "onlineMeetingProvider": "online_meeting_provider",
    "allowNewTimeProposals": "allow_new_time_proposals",
    "occurrenceId": "occurrence_id",
    "isDraft": "is_draft",
    "hideAttendees": "hide_attendees",
    "responseStatus": "response_status",
    "appId": "app_id",
    "passwordCredentials": "password_credentials",
    "iCalUId": "uid",
    "id": "id",
    "body.contentType": "body.content_type",
}


def parse(_dict: dict, scalar_value: bool = False):
    data = None
    try:
        if scalar_value:
            df = pd.DataFrame(_dict, index=[0])
        else:
            df = pd.DataFrame(_dict)
    except ValueError:
        _dict = pd.json_normalize(_dict)
        df = pd.DataFrame(_dict)
    df = df.rename(columns=_map)
    data = df.to_dict("records")
    return data



# Generated by Django 4.2.7 on 2023-12-20 08:47

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbcrm", "0005_account_accountrole_accountroletype_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="activity",
            name="location",
            field=models.CharField(blank=True, max_length=2048, null=True, verbose_name="Location"),
        ),
    ]



# Generated by Django 4.1.8 on 2023-04-25 11:57

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbcrm", "0001_initial_squashed_squashed_0032_productcompanyrelationship_alter_product_prospects_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="activity",
            name="repeat_choice",
            field=models.CharField(
                choices=[
                    ("NEVER", "Never"),
                    ("RRULE:FREQ=DAILY;INTERVAL=1;WKST=MO;BYDAY=MO,TU,WE,TH,FR", "Business Daily"),
                    ("RRULE:FREQ=DAILY", "Daily"),
                    ("RRULE:FREQ=WEEKLY", "Weekly"),
                    ("RRULE:FREQ=WEEKLY;INTERVAL=2", "Bi-Weekly"),
                    ("RRULE:FREQ=MONTHLY", "Monthly"),
                    ("RRULE:FREQ=MONTHLY;INTERVAL=3", "Quarterly"),
                    ("RRULE:FREQ=YEARLY", "Annually"),
                ],
                default="NEVER",
                help_text="Repeat activity at the specified frequency",
                max_length=56,
                verbose_name="Recurrence Frequency",
            ),
        ),
    ]



from django.dispatch import receiver
from wbcore.contrib.directory.viewsets import (
    CompanyModelViewSet,
    EntryModelViewSet,
    PersonModelViewSet,
)
from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig
from wbcore.signals.instance_buttons import add_instance_button


class AccountButtonConfig(ButtonViewConfig):
    def get_custom_instance_buttons(self):
        return {bt.WidgetButton(key="claims", label="Show Claims", icon=WBIcon.TRADE.icon)}

    def get_custom_list_instance_buttons(self):
        return self.get_custom_instance_buttons()


@receiver(add_instance_button, sender=PersonModelViewSet)
@receiver(add_instance_button, sender=EntryModelViewSet)
@receiver(add_instance_button, sender=CompanyModelViewSet)
def crm_adding_instance_buttons(sender, many, *args, **kwargs):
    if many:
        return
    return bt.WidgetButton(key="account", label="Accounts", icon=WBIcon.FOLDERS_MONEY.icon)



from .external_to_internal.create import create_internal_activity_based_on_google_event
from .external_to_internal.delete import (
    delete_recurring_activity,
    delete_single_activity,
)
from .external_to_internal.update import (
    update_activities_from_new_parent,
    update_all_activities,
    update_single_activity,
)
from .internal_to_external.update import (
    update_all_recurring_events_from_new_parent,
    update_all_recurring_events_from_parent,
    update_single_event,
    update_single_recurring_event,
)



import json
import warnings
from datetime import datetime
from typing import Dict

from django.http import HttpRequest, HttpResponse
from django.utils import timezone
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from dynamic_preferences.registries import global_preferences_registry
from google.oauth2.service_account import Credentials
from googleapiclient.discovery import Resource, build
from wbcore.contrib.authentication.models import User
from wbcore.contrib.directory.models import Person

from wbcrm.models import Activity, ActivityParticipant

from .request_utils import (
    create_internal_activity_based_on_google_event,
    delete_recurring_activity,
    delete_single_activity,
    update_activities_from_new_parent,
    update_all_activities,
    update_all_recurring_events_from_new_parent,
    update_all_recurring_events_from_parent,
    update_single_activity,
    update_single_event,
    update_single_recurring_event,
)
from .typing_informations import GoogleEventType
from .utils import GoogleSyncUtils


class GoogleCalendarBackend:
    error_messages = {
        "missing_google_credentials": _(
            "The Google credentials are not set. You cannot use the Google Calendar Backend without the Google credentials."
        ),
        "service_build_error": _("Could not create the Google service. Exception: "),
        "create_error": gettext("Could not create the external google event. Exception: "),
        "delete_error": _("Could not delete a corresponding external event. Exception: "),
        "update_error": gettext("Could not update the external google-event. Exception: "),
        "send_participant_response_error": gettext(
            "Could not update the participation status on the google-event. Exception: "
        ),
        "could_not_sync": _("Couldn't sync with google calendar. Exception:"),
        "could_not_set_webhook": _("Could not set the google web hook for the user: "),
    }

    SCOPES = ["https://www.googleapis.com/auth/calendar"]
    API_SERVICE_NAME = "calendar"
    API_VERSION = "v3"

    @classmethod
    def _get_service_account_file(cls) -> Dict:
        global_preferences = global_preferences_registry.manager()
        google_credentials = global_preferences.get("wbactivity_sync__google_sync_credentials")
        if google_credentials and (serivce_account_file := json.loads(google_credentials)):
            return serivce_account_file
        else:
            raise ValueError(cls.error_messages["missing_google_credentials"])

    @classmethod
    def _get_service_account_url(cls) -> str:
        serivce_account_file = cls._get_service_account_file()
        return serivce_account_file.get("url", "")

    @classmethod
    def _get_service_user_email(cls, activity: Activity) -> str:
        """
        This methods returns the email of the first activity participant with an active google-subscription.
        If no participant with an active subscrition is found, the return value will be an empty string.
        """

        now = timezone.now().replace(tzinfo=None)
        primary_email_contact: str = ""

        def get_email_str(person: Person) -> str:
            if user_profile := User.objects.filter(profile=person).first():
                user_google_backend: Dict = user_profile.metadata.get("google_backend", {})
                expiration: str | None = user_google_backend.get("watch", {}).get("expiration")
                if expiration and datetime.fromtimestamp(int(expiration) / 1000) > now:
                    return str(user_profile.email)
            return ""

        primary_email_contact = get_email_str(activity.creator) if activity.creator else ""
        if not primary_email_contact and (
            internal_participants := activity.participants.filter(
                id__in=Person.objects.filter_only_internal().exclude(
                    id=activity.creator.id if activity.creator else None
                )
            )
        ):
            for participant in internal_participants:
                if primary_email_contact := get_email_str(participant):
                    return primary_email_contact
        return primary_email_contact

    @classmethod
    def _build_service(cls, user_email: str) -> Resource:
        serivce_account_file = cls._get_service_account_file()
        try:
            credentials = Credentials.from_service_account_info(serivce_account_file, scopes=cls.SCOPES)
            return build(cls.API_SERVICE_NAME, cls.API_VERSION, credentials=credentials.with_subject(user_email))
        except Exception as e:
            raise ValueError("{msg}{exception}".format(msg=cls.error_messages["service_build_error"], exception=e))

    @classmethod
    def create_external_activity(cls, activity: Activity) -> None:
        now = timezone.now()
        can_sync_past_activities: bool = global_preferences_registry.manager()["wbactivity_sync__sync_past_activity"]
        if (
            activity.parent_occurrence
            or not (service_user_mail := cls._get_service_user_email(activity))
            or not (service := cls._build_service(user_email=service_user_mail))
            or (not can_sync_past_activities and now > activity.period.lower)  # type: ignore
        ):
            return
        try:
            event_body = GoogleSyncUtils.convert_activity_to_event(activity, True)
            event = service.events().insert(calendarId=service_user_mail, body=event_body).execute()
            metadata = activity.metadata | {"google_backend": {"event": event}}
            Activity.objects.filter(id=activity.id).update(external_id=event["id"], metadata=metadata)
            if Activity.objects.filter(parent_occurrence=activity).exists():
                instances = service.events().instances(calendarId=service_user_mail, eventId=event["id"]).execute()
                google_event_items = instances["items"]
                GoogleSyncUtils.add_instance_metadata(activity, google_event_items, metadata, True)

        except Exception as e:
            raise ValueError("{msg}{exception}".format(msg=cls.error_messages["create_error"], exception=e))

    @classmethod
    def delete_external_activity(cls, activity: Activity) -> None:
        now = timezone.now()
        can_sync_past_activities: bool = global_preferences_registry.manager()["wbactivity_sync__sync_past_activity"]

        if (
            not (service_user_mail := cls._get_service_user_email(activity))
            or not (service := cls._build_service(user_email=service_user_mail))
            or (not can_sync_past_activities and now > activity.period.lower)  # type: ignore
        ):
            return

        try:
            external_id = activity.external_id
            if (
                Activity.objects.filter(parent_occurrence=activity).exists()
                and not not activity.propagate_for_all_children
                and (google_backend := activity.metadata.get("google_backend"))
            ):
                # This step must be done if you want to remove a parent activity without deleting the whole recurring chain.
                # Therefore we use the instance ID instead of the event ID.
                external_id = google_backend.get("instance", {}).get("id")
            service.events().delete(calendarId=service_user_mail, eventId=external_id).execute()
        except Exception as e:
            warnings.warn("{msg}{exception}".format(msg=cls.error_messages["delete_error"], exception=e))

    @classmethod
    def update_external_activity(cls, activity: Activity) -> None:
        if not activity.metadata.get("google_backend"):
            cls.create_external_activity(activity)
            activity.refresh_from_db()

        if not (service_user_mail := cls._get_service_user_email(activity)) or not (
            service := cls._build_service(user_email=service_user_mail)
        ):
            return

        updated_event_body = GoogleSyncUtils.convert_activity_to_event(activity)
        try:
            is_parent = Activity.objects.filter(parent_occurrence=activity).exists()

            def update_all_recurring_events():
                if activity.metadata.get("old_parent_id"):
                    update_all_recurring_events_from_new_parent(
                        service_user_mail, service, activity, updated_event_body
                    )
                else:
                    update_all_recurring_events_from_parent(service_user_mail, service, activity, updated_event_body)

            if is_parent or activity.parent_occurrence:
                if activity.propagate_for_all_children:
                    update_all_recurring_events()
                else:
                    update_single_recurring_event(service_user_mail, service, activity, updated_event_body)
            else:
                update_single_event(service_user_mail, service, activity, updated_event_body)
        except Exception as e:
            raise ValueError("{msg}{exception}".format(msg=cls.error_messages["update_error"], exception=e))

    @classmethod
    def send_participant_response_external_activity(
        cls, activity_participant: ActivityParticipant, response_status: str
    ):
        participant: Person | None = Person.objects.filter(id=activity_participant.participant.id).first()
        activity: Activity | None = Activity.objects.filter(id=activity_participant.activity.id).first()
        if not participant or not activity:
            return
        if Activity.objects.filter(parent_occurrence=activity).exists():
            google_backend = activity.metadata.get("google_backend", {})
            external_id: str | None = google_backend.get("instance", google_backend.get("event", {})).get("id", None)
        else:
            external_id: str | None = activity.external_id

        creator_mail = str(activity.creator.primary_email_contact()) if activity.creator else ""
        participant_mail = str(participant.primary_email_contact())
        service: Resource = cls._build_service(user_email=creator_mail)

        if not service or not external_id:
            return
        try:
            google_status = GoogleSyncUtils.convert_participant_status_to_attendee_status(response_status)
            instance: Dict = service.events().get(calendarId=creator_mail, eventId=external_id).execute()
            attendees_list: list[Dict] = instance.get("attendees", [])

            for index, attendee in enumerate(attendees_list):
                if attendee.get("email") == participant_mail:
                    attendees_list[index]["responseStatus"] = google_status
                    break
            metadata = activity.metadata
            google_backend = metadata.get("google_backend", {})
            event_metadata = google_backend.get("event", google_backend.get("instance", {"instance": {}}))
            event_metadata |= instance
            Activity.objects.filter(id=activity.id).update(metadata=metadata)

            service.events().patch(calendarId=creator_mail, eventId=instance["id"], body=instance).execute()

        except Exception as e:
            raise ValueError(
                "{msg}{exception}".format(msg=cls.error_messages["send_participant_response_error"], exception=e)
            )

    @classmethod
    def sync_with_external_calendar(cls, request: HttpRequest) -> HttpResponse:
        if (
            request.headers
            and (channel_id := request.headers.get("X-Goog-Channel-Id"))
            and User.objects.filter(pk=channel_id).exists()
        ):
            pass  # TODO handle_changes_as_task.delay(channel_id)
        return HttpResponse({})

    @classmethod
    def get_sync_token(cls, user: User) -> str | None:
        if google_backend := user.metadata.get("google_backend"):
            return google_backend.get("sync_token")

    @classmethod
    def delete_internal_activity(cls, activity: Activity, **kwargs) -> None:
        event = kwargs.get("event", {})
        user_email = kwargs.get("user_email", "")
        service = kwargs.get("service")
        if Activity.objects.filter(parent_occurrence=activity).exists() or activity.parent_occurrence:
            delete_recurring_activity(activity, event, user_email, service)
        else:
            delete_single_activity(activity)

    @classmethod
    def update_internal_activity(cls, activity: Activity, **kwargs) -> None:
        event: GoogleEventType = kwargs.get("event", {})
        user_email = kwargs.get("user_email", "")
        service = kwargs.get("service")
        if event.get("recurringEventId"):
            update_single_activity(event, activity)
        elif event.get("recurrence"):
            update_all_activities(activity, event, user_email, service)
        else:
            update_single_activity(event, activity)

    @classmethod
    def create_internal_activity(cls, **kwargs) -> None:
        event = kwargs.get("event", {})
        user = kwargs.get("user")
        service = kwargs.get("service")
        create_internal_activity_based_on_google_event.si(event, user, service)

    @classmethod
    def handle_changes(cls, user_id: int) -> None:
        user = User.objects.get(id=user_id)
        user_email = user.email
        can_sync_past_activities: bool = global_preferences_registry.manager()["wbactivity_sync__sync_past_activity"]
        service = cls._build_service(user_email=user_email)
        now = timezone.now()
        if not service:
            return
        external_event_list = []
        page_token = None
        while True:
            request = service.events().list(
                calendarId=user_email, pageToken=page_token, syncToken=cls.get_sync_token(user)
            )
            events = {}
            try:
                events: Dict = request.execute()
            except Exception as e:
                warnings.warn("{msg}{exception}".format(msg=cls.error_messages["could_not_sync"], exception=e))
            external_event_list += events.get("items", [])
            page_token = events.get("nextPageToken")
            if not page_token:
                user.metadata.setdefault("google_backend", {})
                user.metadata["google_backend"]["sync_token"] = events.get("nextSyncToken")
                user.save()
                break

        for event in external_event_list:
            if event.get("start") and not event.get("recurrence"):
                event_start, _ = GoogleSyncUtils.get_start_and_end(event)
                is_all_day_event = True if event["start"].get("date") else False
                starts_in_past = now.date() > event_start.date() if is_all_day_event else now > event_start
                if not can_sync_past_activities and starts_in_past:
                    return
            external_id = event["id"]
            # Note about how Google assigns IDs for events: A single, non-recurring event has an ID consisting of a unique string.
            # As soon as an event has recurring subsequent events, this string is extended by the start date of the respective subsequent event. these two parts are connected by "_R".
            # So if you look at the part before _R you get the ID for the parent event.
            first_part_of_id = external_id.split("_R")[0] if "_R" in external_id else None
            if (
                (activity := Activity.objects.filter(external_id=external_id).first())
                or (activity := Activity.objects.filter(metadata__google_backend__instance__id=external_id).first())
                or (activity := Activity.objects.filter(metadata__google_backend__event__id=external_id).first())
            ):
                # There are two ways we know an event was deleted. Either we receive the event-status "cancelled", or when the "recurrence" field changes.
                # The second one can also indicate that an event was altered. But at this point we don't know if it was deleted or updated. If it was updated we can restore it later.
                google_backend: Dict = activity.metadata["google_backend"]
                metadata_event: Dict = google_backend.get("event", google_backend.get("instance", {}))
                metadata_event_reccurence: list[str] | None = metadata_event.get("recurrence")

                if event.get("status") == "cancelled" or (
                    event.get("recurrence") and event.get("recurrence") != metadata_event_reccurence
                ):
                    cls.delete_internal_activity(activity, event=event, user_email=user_email, service=service)
                else:
                    cls.update_internal_activity(activity, event=event, user_email=user_email, service=service)
            elif first_part_of_id and (
                parent_occurrence := Activity.objects.filter(external_id=first_part_of_id).first()
            ):
                update_activities_from_new_parent(event, parent_occurrence, user_email, service)
            else:
                cls.create_internal_activity(event=event, user=user, service=service)

    @classmethod
    def get_external_activity(cls, activity: Activity):
        pass

    @classmethod
    def forward_external_activity(cls, activity: Activity, participants: list):
        cls.update_external_activity(activity)

    @classmethod
    def set_web_hook(cls, user: User, expiration_in_ms: int = 0) -> None:
        user_email = user.email
        service = cls._build_service(user_email=user_email)
        if service:
            try:
                watch_body = {
                    "id": user.id,  # type: ignore
                    "type": "web_hook",
                    "address": cls._get_service_account_url(),
                }
                if expiration_in_ms > 0.0:
                    watch_body |= {"expiration": str(expiration_in_ms)}
                response = service.events().watch(calendarId=user_email, body=watch_body).execute()
                user.metadata.setdefault("google_backend", {})
                user.metadata["google_backend"]["watch"] = response
                user.save()
            except Exception as e:
                raise ValueError(
                    "{msg}{user}. Eception: {exception}".format(
                        msg=cls.error_messages["could_not_set_webhook"],
                        user=user.profile.computed_str,
                        exception=e,  # type: ignore
                    )
                )

    @classmethod
    def stop_web_hook(cls, user: User) -> None:
        user_email = user.email
        service = cls._build_service(user_email=user_email)
        if service:
            body = {
                "id": user.metadata["google_backend"]["watch"]["id"],
                "resourceId": user.metadata["google_backend"]["watch"]["resourceId"],
            }
            service.channels().stop(body=body).execute()
            del user.metadata["google_backend"]["watch"]
            user.save()

    @classmethod
    def check_web_hook(cls, user: User) -> None:
        now = timezone.now().replace(tzinfo=None)
        user_google_backend: Dict = user.metadata.get("google_backend", {})
        expiration: str | None = user_google_backend.get("watch", {}).get("expiration")
        if expiration and datetime.fromtimestamp(int(expiration) / 1000) > now:
            warnings.warn(_("Timestamp valid until:") + str(datetime.fromtimestamp(int(expiration) / 1000)))
        else:
            raise Exception(_("No valid web hook found"))

    def _get_webhook_inconsistencies(self) -> str: ...

    def webhook_resubscription(self) -> None: ...
